/*

  barrier.c: barrier functions of the psthreads library. 
  Copyright (C) 2002 Thomas Repantis <darkzero@otenet.gr>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/
 
/*
  barrier.c: barrier functions of the psthreads library. 
*/
 
/*

  These functions implement the Sense Reversing Centralized Incremental Barrier.

*/

#include <stdio.h>
#include "machine.h"
#include "barrier.h"

#define UNFREE 0  /* Not-free, you cannot leave the barrier. */
#define FREE   1  /* Free, you can leave the barrier. */



void _ps_barrier_init(psthread_barrier_t *thr_incr_barrier, int processors)
{
  /* incr_barrier_init initializes the incr_barrier. */
  thr_incr_barrier->global_sense = UNFREE;
  thr_incr_barrier->processors = processors;
  thr_incr_barrier->counter = processors;
}

void _ps_barrier(psthread_barrier_t *thr_incr_barrier)
{
  /* incr_barrier makes the thread wait until all threads reach the incr_barrier pointed. */
 
  int local_sense;

  local_sense = thr_incr_barrier->global_sense;

  /* The if control must take place together with the fetch and add (atomically). */
  if( fetch_and_add(&thr_incr_barrier->counter, -1) != 1)   /* I am not the last one reaching the barrier. */
    {
      while( thr_incr_barrier->global_sense == local_sense )
	;
    }
  else  /* thr_incr_barrier->counter == 1 */   /* I am the last one reaching the barrier. */
    {
      thr_incr_barrier->counter = thr_incr_barrier->processors; 
      thr_incr_barrier->global_sense = ~thr_incr_barrier->global_sense; /* Bit negation or normal one (using !). */
    }

}


int psthread_barrier_init(psthread_barrier_t *barrier, int thread_count)
{
  /* incr_barrier_init initializes the incr_barrier. */
  barrier->global_sense = UNFREE;
  barrier->processors = thread_count;
  barrier->counter = thread_count;
  return 0;
}

int psthread_barrier_wait(psthread_barrier_t *barrier)
{
  /* incr_barrier makes the thread wait until all threads reach the incr_barrier pointed. */
 
  int local_sense;

  local_sense = barrier->global_sense;

  /* The if control must take place together with the fetch and add (atomically). */
  if( fetch_and_add(&barrier->counter, -1) != 1)   /* I am not the last one reaching the barrier. */
    {
      while( barrier->global_sense == local_sense )
	psthread_yield();
    }
  else  /* barrier->counter == 1 */   /* I am the last one reaching the barrier. */
    {
      barrier->counter = barrier->processors; 
      barrier->global_sense = ~barrier->global_sense; /* Bit negation or normal one (using !). */
      /* Now all others can wake up... */
    }
  return 0;
}


