/*

  libpsthreads.h: Header file to use the psthreads library.
  Copyright (C) 2002 Thomas Repantis <darkzero@otenet.gr>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/
 
/*

  This header file contains the declarations needed to use the psthreads library. 

*/


#include <setjmp.h>
#include <pthread.h>
#include <errno.h>

#include "stack.h"
#include "lock.h"
#include "barrier.h"

/* Definitions of a thread's status. */

#define RUNNING 0
#define READY   1
#define ZOMBIE  2


int NUM_OF_VPs; /* The number of Virtual Processors and of the kernel-level 
		   threads (maximally equal to the number of processors). */

/* This is the descriptor of a user-level thread. */

typedef struct _ps_thread {
  ps_stack_t *stack;         /* Pointer to the thread's stack. */
  struct _ps_thread *prev;   /* Pointer to the previous thread. */
  struct _ps_thread *next;   /* Pointer to the next thread. */
  jmp_buf context;           /* Space to store the thread's environment. */
  int status;                /* Thread's status (running, ready, zombie). */
  void (*start_routine)(void *); /* Thread's start routine. */
  void *arg;                     /* Thread's start routine's arguments. */
  struct _ps_thread *whojoinedthisthr; /* Struct to help interaction of 
					  psthread_exit() and psthread_join(). 
					  Pointer to the thread who joined this 
					  thread (if any). */
  psthread_lock_t tlock;     /* Lock to synchronize changes to the thread - exits and joins. */
} psthread_t;


/* This is the descriptor of a queue. */
/* The queue is implemented as a circular doubly-linked list, protected by a lock. */

typedef struct _ps_queue {
  psthread_lock_t qlock; 
  psthread_t *head;
} ps_queue_t;

extern void _ps_queue_init(ps_queue_t *q);
extern void _ps_queue_insert(ps_queue_t *q, psthread_t *t);
extern void _ps_queue_append(ps_queue_t *q, psthread_t *t);
extern psthread_t * _ps_queue_removefirst(ps_queue_t *q);
extern psthread_t * _ps_queue_remove(ps_queue_t *q, psthread_t *t);


ps_queue_t ready_queue;
ps_queue_t zombie_queue;

/* Kernel-level thread for each VP. */
pthread_t *_ps_klt;

/* Pointer to the idle user-level thread for each Virtual Processor. */
psthread_t **_ps_idle_thread;   

/* Pointer to the currently active user-level thread for each Virtual Processor. */
psthread_t **_ps_active_thread; 

volatile long int _ps_actthr_counter;  /* Counter of the active user-level 
					  threads, to determine the end of 
					  the execution.  Declared as volatile, 
				          since many threads can change it.*/

psthread_lock_t _mutex_counter;        /* To protect the _ps_actthr_counter. */


extern int psthread_init(int kernel_threads);
extern int psthread_create(psthread_t **thread_desc, void (start_routine)(void *), void *arg, int queue);
extern void psthread_exit(void);
extern int psthread_join(psthread_t *thread_desc);
extern void psthread_yield(void);
extern psthread_t psthread_self(void);

extern void _ps_run(void);
extern int _ps_find_vpnum(void);
extern void _ps_idle(void);
extern void _ps_schedule(void);
extern void *_ps_vp_init(void *vpnumber);
