/*

  matmul.c: Matrix Multiplication (Inner Product) using the psthreads library 
  Copyright (C) 2002 Thomas Repantis <darkzero@otenet.gr>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/
 
/*
  matmul.c: Matrix Multiplication (Inner Product) using the psthreads library. 
*/


#include <stdio.h>
#include <stdlib.h>       
#include <sys/time.h>
#include <errno.h>    
#include <math.h>
#include "libpsthreads.h"

#define MICROS_PER_SECOND 1000000.0
#define OFF 0
#define ON  1
#define N   2048          /* The size of the matrix. */


double (*a)[N], (*b)[N], (*c)[N], (*c_seq)[N];


long int klt_num;
long int ult_num;


void matr_init()
{
  int i,j;
  double swap;

  /* Fill the matrices to be multiplied with random numbers, 
     which have a decimal part. */
  fprintf(stdout, "Filling the matrices with random numbers...\n");
  fflush(stdout);

  for (i=0;i<N;i++) 
    for (j=0;j<N;j++){
      a[i][j]=( (double)random()/(double)RAND_MAX );
      b[i][j]=( (double)random()/(double)RAND_MAX );
    }
  
  /* Find the transpose of matrix b to optimize the multiplication
     by using locality. */
  for (i=0;i<N;i++)
    for (j=i+1;j<N;j++){
      swap=b[i][j];
      b[i][j]=b[j][i];
      b[j][i]=swap;
    }
  fprintf(stdout, "Finished the initialization of the matrices.\n");
  fflush(stdout);
}


void matr_multiply(void *arg)
{
  /* The argument is the id of the user-level thread. */
  
  int i,j,k;
  int start, end;
  double temp;
  
  /* Produce some output (so that the user sees the progress).  
     When time is measured, fprintf() is commented out. */
  /* Since psthread_self() returns -according to the API- a psthread_t structure 
     (and not a pointer to it), what we print with %d is just an id without a 
     particular meaning. */
  //    fprintf(stdout, "ULT number %d with descriptor %d started on VP %d.\n", 
  //	  (int)arg, psthread_self(), _ps_find_vpnum());
  //    fflush(stdout);
  
  
  if( N%ult_num == 0 )
    start=(N/ult_num)*(int)arg;
  else
    start=( (N/ult_num)+1 )*(int)arg;
  
  if( N%ult_num == 0 )
    end=start+(N/ult_num);
  else
    end=start+(N/ult_num)+1;

  /* We don't want to move outside the matrix and since there are several 
     combinations of N, ult_num and klt_num that could result to that, we 
     just do a separate check... */  
  if( start < 0 )
    start = 0;
  if( end > N )
    end = N;
  
  for (j=0;j<N;j++){ 
    for (i=start;i<end;i++){
      temp=0.0;
      for (k=0;k<N;k++)
	/* Since we have found the transpose of b we can compute 
	   temp+=a[i][k]*b[j][k], instead of temp+=a[i][k]*b[k][j], 
	   improving the performance dramatically (~5 times faster). */
	temp+=a[i][k]*b[j][k];
      c[i][j]=temp;
      /* Produce some output (so that the user sees the progress), 
	 but do not show all of the product matrix elements (since 
	 they are too many), only the first of each column. */
      /* When time is measured, printf() is commented out. */
      //      if (i==start)
      //	printf("c[%d][%d]=%f\n",i,j,c[i][j]);
    }
  }
  /* This ult finished its job and exits. */
  psthread_exit();
}

void matr_multiply_seq(void)
{
  int i,j,k;
  int start, end;
  double temp;
  
  start=0;
  end=N;
  
  for (j=0;j<N;j++){ 
    for (i=start;i<end;i++){
      temp=0.0;
      for (k=0;k<N;k++)
	/* Since we have found the transpose of b we can compute 
	   temp+=a[i][k]*b[j][k], instead of temp+=a[i][k]*b[k][j], 
	   improving the performance dramatically (~5 times faster). */
	temp+=a[i][k]*b[j][k];
      c_seq[i][j]=temp;
      /* Produce some output (so that the user sees the progress), 
	 but do not show all of the product matrix elements (since 
	 they are too many), only the first of each column. */
      /* When time is measured, printf() is commented out. */
      //      if (i==start)
      //	printf("c_seq[%d][%d]=%f\n",i,j,c_seq[i][j]);
    }
  }
}


/* Compare the resulting matrices, the one produced using the psthreads library with the one 
   produced sequentially. */
void compare_results()
{
  int i,j;

  for(i=0;i<N;i++)
    for(j=0;j<N;j++)
      /* Test the absolute value of the difference. */
      if( fabs( (c[i][j] - c_seq[i][j]) ) >= 0.001 ){
	printf("The result does not match the sequential result at least in 3 decimal parts.\n");    
	return;
      }
  printf("The result matches the sequential result at least in 3 decimal parts.\n");    
  
}


int main(int argc, char *argv[])
{
  float time_elapsed;
  struct timeval val1, val2;
  float time_elapsed_seq;
  struct timeval val1_seq, val2_seq;

  int i;
  int test = OFF;        /* Flag to turn testing on or off. 
			    On only if the user says so. */


  psthread_t **ult_descr;

  /*----------------------------------------------*/
  /* Introduce yourself. */
  /*----------------------------------------------*/

  printf("\n");
  printf("matmul.c: Matrix Multiplication (Inner Product) using the psthreads library. \n");  
  printf("Copyright (C) 2002 Thomas Repantis <darkzero@otenet.gr>\n");
  printf("This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you are welcome to redistribute it under certain conditions. matmul is licensed under the GNU General Public License. See the file COPYING for details.\n");
  printf("\n");

  if( argc != 3 && argc != 4)
    {
      printf("Usage: matmul [Number of virtual processors (klts) to be used] [Number of user-level threads to be used] [-t]\n");
      printf("-t to optionally test the result sequentially.\n");
      printf("For example: matmul 4 100 -t\n");
      printf("\n");
      exit(1);
    }
  
  /*----------------------------------------------*/
  /* Convert the command line argument strings to unsigned long integers. */
  /*----------------------------------------------*/
  
  errno = 0;
  klt_num = strtoul(argv[1], (char**)NULL, 10);
  if( errno != 0 )
    {
      fprintf(stderr, "Error on the number of virtual processors (%s) given.\n", argv[1]);
      fflush(stderr);
      exit(1);
    }

  errno = 0;
  ult_num = strtoul(argv[2], (char**)NULL, 10);
  if( errno != 0 )
    {
      fprintf(stderr, "Error on the number of user-level threads (%s) given.\n", argv[2]);
      fflush(stderr);
      exit(1);
    }
  
  if( argc == 4 ){
    if( strcmp(argv[3], "-t") == 0) 
      test = ON;
    else{
      fprintf(stderr, "Error on the parameter (%s) given.\n", argv[3]);
      fflush(stderr);
      exit(1);
    }
  }
      
  fprintf(stdout, "The number of virtual processors given is interpreted as: %ld\n", klt_num);
  fflush(stdout);
  
  fprintf(stdout, "The number of user-level threads given is interpreted as: %ld\n", ult_num);
  fflush(stdout);
  
  if( test == ON){
    fprintf(stdout, "Sequential testing is turned on.\n\n");
    fflush(stdout);
  }
  
  /*----------------------------------------------*/
  /* Check for valid input. */
  /*----------------------------------------------*/

  if( klt_num == 0 || klt_num<0 )
    {
      fprintf(stderr, "Error on the number of virtual processors given (%ld). \n", klt_num);
      fprintf(stderr, "There must at least be one virtual processor.\n");
      fflush(stderr);
      exit(1);
    }

  if( ult_num == 0 || ult_num<0 )
    {
      fprintf(stderr, "Error on the number of user-level threads given (%ld). \n", ult_num);
      fprintf(stderr, "There must at least be one user-level thread.\n");
      fflush(stderr);
      exit(1);
    }
  
  /* We allow the user to give values that may not make sense, like klt_num > ult_num 
     or ult_num>N, since they are just inefficient and do not harm otherwise. */ 

  /*----------------------------------------------*/
  /* Memory allocation. */                  
  /*----------------------------------------------*/


  if ( (ult_descr = (psthread_t **)calloc(ult_num, sizeof(psthread_t *)))==NULL )
    return -1;

  /* Since we know the matrix size we could also use static memory allocation.  
     We use dynamic, in case we would like the size to be user-defined. */

  if ( (a = (double (*)[N])calloc(N*N, sizeof(double)))==NULL )
    return -1;
  if ( (b = (double (*)[N])calloc(N*N, sizeof(double)))==NULL )
    return -1;
  if ( (c = (double (*)[N])calloc(N*N, sizeof(double)))==NULL )
    return -1;
  if ( (c_seq = (double (*)[N])calloc(N*N, sizeof(double)))==NULL )
    return -1;

  /*----------------------------------------------*/
  /* Initialize the matrices to be multiplied. */
  /*----------------------------------------------*/
  matr_init();

  fprintf(stdout, "Mutliplying using the psthreads library...\n");
  fflush(stdout);
  
  /* Note the time the multiplication started. */
  gettimeofday(&val1,NULL);
  
  /*----------------------------------------------*/
  /* Initialize the psthreads library. */
  /*----------------------------------------------*/
  psthread_init(klt_num);
  
  /*----------------------------------------------*/
  /* Create the ults and do the multiplication. */
  /*----------------------------------------------*/
  for(i=0;i<ult_num;i++){
    psthread_create(&ult_descr[i], matr_multiply, (void *)i, -1 );
  }

  /* So that vp 0 can execute ults too, we must call psthread_yield() or 
     psthread_join().  If we want the execution to start early, we call 
     psthread_yield() before the several calls to psthread_join(). */
  //psthread_yield();  

  /*----------------------------------------------*/
  /* Wait for the ults to finish. */
  /*----------------------------------------------*/
  for(i=0;i<ult_num;i++){
    psthread_join(ult_descr[i]);
  }    

  /* Note the time the multiplication finished. */
  gettimeofday(&val2,NULL);
    
  /* Calculate the seconds part of the time elapsed. */
  time_elapsed = (val2.tv_sec - val1.tv_sec)*MICROS_PER_SECOND;
  
  /* Calculate the microseconds part of the time elapsed. */
  time_elapsed += (val2.tv_usec - val1.tv_usec);
  
  /* Convert the time elapsed in seconds. */
  time_elapsed = time_elapsed  / MICROS_PER_SECOND;
  
  printf("Total time needed for matrix multiplication using the psthreads library:  %10.2f seconds\n", time_elapsed);
  

  /*----------------------------------------------*/
  /* Sequential testing. */
  /*----------------------------------------------*/
  if( test == ON){

    fprintf(stdout, "Mutliplying sequentially to test the results...\n");
    fflush(stdout);

    /* Note the time the multiplication started. */
    gettimeofday(&val1_seq,NULL);
    matr_multiply_seq();
    /* Note the time the multiplication finished. */
    gettimeofday(&val2_seq,NULL);
  
    /* Calculate the seconds part of the time elapsed. */
    time_elapsed_seq = (val2_seq.tv_sec - val1_seq.tv_sec)*MICROS_PER_SECOND;
    /* Calculate the microseconds part of the time elapsed. */
    time_elapsed_seq += (val2_seq.tv_usec - val1_seq.tv_usec);
    /* Convert the time elapsed in seconds. */
    time_elapsed_seq = time_elapsed_seq  / MICROS_PER_SECOND;
  
    printf("Total time needed for sequential matrix multiplication:  %10.2f seconds\n", 
	   time_elapsed_seq);
    
    compare_results();
    
  }
    

 /*----------------------------------------------*/
  /* Clean up. */
  /*----------------------------------------------*/
  free(ult_descr);
  free(a);
  free(b);
  free(c);

  return 0; 
}
