/*

  thrmanipul.c: The part of the psthreads non-preemptive, user-level thread 
  library concerning the manipulation of the threads.
  Copyright (C) 2002 Thomas Repantis <darkzero@otenet.gr>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/
 
/*
  
  These functions of the psthreads library concern the
  manipulation of the threads.

*/

#include <stdio.h>
#include <stdlib.h>
#include "libpsthreads.h"
  
int psthread_init(int kernel_threads)
{
  psthread_t *dummy_thread;
  int i;
  pthread_t kltid;
  pthread_attr_t klt_attribs;


  /* Initialize internal structures. */

  NUM_OF_VPs = kernel_threads;
  _ps_actthr_counter = 0;

  if ( (_ps_klt = (pthread_t *)calloc(NUM_OF_VPs, sizeof(pthread_t)))==NULL )
    return EAGAIN;
  if ( (_ps_idle_thread = (psthread_t **)calloc(NUM_OF_VPs, sizeof(psthread_t *)))==NULL )
    return EAGAIN;
  if ( (_ps_active_thread = (psthread_t **)calloc(NUM_OF_VPs, sizeof(psthread_t *)))==NULL )
    return EAGAIN;

  _ps_lock_init( &_mutex_counter );

  _ps_queue_init(&ready_queue);
  _ps_queue_init(&zombie_queue);

  /* Store the kernel-level thread for VP 0. */
  _ps_klt[0]=pthread_self();

  /* Create idle user-level thread for VP 0. */
  if( psthread_create(&_ps_idle_thread[0], (void *)_ps_idle, NULL, -1) != 0)
    return EAGAIN; 

  /* Create dummy user-level thread corresponding to the code already
     being executed. We need this ult just for the first context-switch. 
     We do not use psthread_create, as we do not want this ult to be 
     appended to the ready-queue. */
  dummy_thread = (psthread_t *)calloc(1, sizeof(psthread_t));
  if ( dummy_thread == NULL )
    return EAGAIN;
  
  dummy_thread->status = RUNNING;
  _ps_active_thread[0] = dummy_thread;
  _ps_lock( &_mutex_counter );
  _ps_actthr_counter++;
  _ps_unlock( &_mutex_counter );

  /* Create kernel_threads-1 more kernel-level threads (for the rest of 
     the vps, since we (vp 0) already have one. We use pthread_create 
     for POSIX threads (or alternatively Solaris's thr_create()). 
     We create threads with default stack and stack size and make them 
     execute _ps_vp_init for each VP. The argument of the _ps_vp_init 
     is the vpnum. The threads are bound (scheduled on a system-wide 
     basis) and the actual id of each one is returned to kltid. */
  
  pthread_attr_init(&klt_attribs);
  pthread_attr_setscope(&klt_attribs, PTHREAD_SCOPE_SYSTEM);
  
  for( i=1; i<kernel_threads; i++ ){
    if( pthread_create( &kltid, &klt_attribs, _ps_vp_init, (void *)i ) != 0 ){
      perror("Thread creation error.");
      return EAGAIN;
    }
  }
  
  return 0;
}


int psthread_create(psthread_t **thread_desc, void (start_routine)(void *), void *arg, int queue)
{
  *thread_desc = (psthread_t *)calloc(1, sizeof(psthread_t));
  if( thread_desc == NULL )
    return EAGAIN;
  
  (*thread_desc)->stack = _ps_stack_new();
  if( (*thread_desc)->stack == NULL)
    return EAGAIN;
  
  (*thread_desc)->status = READY;
  (*thread_desc)->start_routine = start_routine;
  (*thread_desc)->arg = arg;
  (*thread_desc)->whojoinedthisthr = NULL;

  MD_INIT_CONTEXT(*thread_desc, (*thread_desc)->stack->sp, _ps_run);
  
  _ps_lock_init( &((*thread_desc)->tlock) );

  if (queue!= -1)
    return EAGAIN;
  
  _ps_queue_append( &ready_queue, *thread_desc );

  _ps_lock( &_mutex_counter );
  _ps_actthr_counter++;
  _ps_unlock( &_mutex_counter );

  return 0;
					
}

void psthread_exit(void)
{
  /* Find the active thread on the VP we are. */
  int vpnum = _ps_find_vpnum();
  psthread_t *actthr;
  actthr=_ps_active_thread[vpnum];

  _ps_lock( &(actthr->tlock) );
  
  actthr->status = ZOMBIE;
  _ps_queue_append( &zombie_queue, actthr );

  if( actthr->whojoinedthisthr != NULL ){
    actthr->whojoinedthisthr->status = READY;
    _ps_queue_insert( &ready_queue, actthr->whojoinedthisthr );
  }
   
  _ps_lock( &_mutex_counter );
  _ps_actthr_counter--;
  _ps_unlock( &_mutex_counter );
  
  _ps_unlock( &(actthr->tlock) );
  
  _ps_schedule();

}

int psthread_join(psthread_t *thread_desc)
{
  /* Find the active thread on the VP we are. */
  int vpnum = _ps_find_vpnum();
  psthread_t *actthr;
  actthr=_ps_active_thread[vpnum];
  
  _ps_lock(&thread_desc->tlock);  
  
  if( thread_desc == NULL )
    return ESRCH;
    
  if( thread_desc->whojoinedthisthr != NULL )
    return EINVAL;
    
  if( actthr == thread_desc )
    return EDEADLK;

  if( thread_desc->status == ZOMBIE ){
    if(_ps_queue_remove( &zombie_queue, thread_desc ) != thread_desc){
      _ps_unlock( &(thread_desc->tlock) );
      return ESRCH;
    }
    else{
      free( thread_desc->stack->vaddr );
      free( thread_desc );
      /* After freeing there is no need to unlock... */
      /* We needed the lock to synchronize exit and join, 
	 so that we wouldn't free memory too early. */ 
      return 0;
    }
  }
  else{
    thread_desc->whojoinedthisthr = actthr;
    _ps_unlock(&thread_desc->tlock);
    _ps_schedule();
    return 0;
  }
}

void psthread_yield(void)
{
  /* Find the active thread on the VP we are. */
  int vpnum = _ps_find_vpnum();
  psthread_t *actthr;
  actthr = _ps_active_thread[vpnum];

  actthr->status = READY;
  _ps_queue_append( &ready_queue, actthr );

  _ps_schedule();
}

psthread_t psthread_self(void)
{
  /* Find the active thread on the VP we are. */
  int vpnum = _ps_find_vpnum();
  psthread_t *actthr;
  actthr=_ps_active_thread[vpnum];

  return *actthr;
}



